(* $Id: godi_plist.ml 494 2004-09-30 13:00:07Z gerd $
 * ----------------------------------------------------------------------
 * Copyright 2004 by Gerd Stolpmann.
 *
 * This file is part of godi-tools.
 *
 * godi-tools is free software; you can redistribute it and/or modify it
 * under the conditions in the file LICENSE.
 *
 * godi-tools is distributed WITHOUT ANY WARRANTY, as detailed in the
 * file LICENSE.
 *)

open Godi_versions
open Printf

type inst_plist_element =
    [ `Name of string
    | `Cwd of string
    | `Src of string
    | `Pkgdep of Godi_versions.dependency
    | `Blddep of Godi_versions.dependency
    | `Pkgcfl of Godi_versions.dependency
    | `Dirrm of string
    | `File of string * string option
    | `Comment of string
    | `Exec of string
    | `Unexec of string
    | `Display of string
    | `Option of [ `Extract_in_place | `Preserve ]
    | `Ignore
    | `Unsupported of string * string
    ]

exception Plist_error of string * string

let word_re = Str.regexp "[\r\n\t ]+";;

let parse_inst_plist_verbatim ?(no_files = false) file next_line =
  (* Parses the file, but `File elements do not include checksums.
   * The checksum are represented as in the file, i.e. as comments.
   *)
  let rec parse() =
    try
      let line = next_line () in  (* or End_of_file *)
      let words = Str.split word_re line in
      let butfirst_word() =
	match Str.bounded_split word_re line 2 with
	    [] | [ _ ] -> ""
	  | [ _; word ] -> word
	  | _ -> assert false
      in
      let parse_dep s =
	try parse_dependency s
	with Failure _ ->
	  raise(Plist_error(file, line))
      in
      ( match words with
	    [ "@name"; name ] -> `Name name :: parse ()
	  | [ ( "@cwd" | "@cd"); dir ] -> `Cwd dir :: parse ()
	  | [ "@src"; dir ] -> `Src dir :: parse ()
	  | [ "@pkgdep"; s ] -> `Pkgdep(parse_dep s) :: parse ()
	  | [ "@blddep"; s ] -> `Blddep(parse_dep s) :: parse ()
	  | [ "@pkgcfl"; s ] -> `Pkgcfl(parse_dep s) :: parse ()
	  | [ "@dirrm"; dir ] ->
	      if no_files then parse () else `Dirrm dir :: parse ()
	  | "@comment" :: _ ->
	      if no_files then parse () else
		let rest = butfirst_word() in `Comment rest :: parse ()
	  | "@exec" :: _ ->
	      let rest = butfirst_word() in `Exec rest :: parse ()
	  | "@unexec" :: _ ->
	      if no_files then parse () else
		let rest = butfirst_word() in `Unexec rest :: parse ()
	  | [ "@display"; file ] ->
	      if no_files then parse () else
		`Display file :: parse ()
	  | [ "@option"; "extract-in-place" ] ->
	      `Option `Extract_in_place :: parse ()
	  | [ "@option"; "preserve" ] ->
	      `Option `Preserve :: parse ()
	  | [ "@ignore" ] ->
	      `Ignore :: parse()
	  | word :: _ when word <> "" && word.[0] = '@' ->
	      if no_files then parse () else
		`Unsupported(List.hd words, butfirst_word()) :: parse ()
	  | [ file ] ->
	      if no_files then parse () else
		`File(line,None) :: parse ()
	  | [] -> parse ()
	  | _ -> raise(Plist_error(file, line))
      )
    with
	End_of_file -> []
  in

  parse()
;;


let from_file parse_fun file =
  let f = open_in file in
  let next_line() = input_line f in
  try
    let r = parse_fun next_line in
    close_in f;
    r
  with error ->
    close_in f;
    raise error
;;


let from_string parse_fun s =
  let lines = ref (Str.split_delim (Str.regexp "\n") s) in
  let next_line() =
    match !lines with
	[] -> raise End_of_file
      | line :: lines' ->
	  lines := lines';
	  line
  in
  parse_fun next_line
;;


let md5_syntax_re = Str.regexp "^MD5:[0-9a-f]+$" ;;

let parse_inst_plist_from ?no_files file next_line =
  (* Also parse MD5 checksums: *)

  let md5_syntax s =
    Str.string_match md5_syntax_re s 0 in

  let rec incorporate_checksums elements =
    match elements with
	`File(file,None) :: `Comment c :: rest when md5_syntax c ->
	  `File(file, Some c) :: incorporate_checksums rest
      | e :: rest ->
	  e :: incorporate_checksums rest
      | [] ->
	  []
  in

  incorporate_checksums(parse_inst_plist_verbatim ?no_files file next_line)
;;


let parse_inst_plist ?no_files file =
  from_file (parse_inst_plist_from ?no_files file) file ;;


let parse_inst_plist_string ?no_files str =
  from_string (parse_inst_plist_from ?no_files "<nofile>") str ;;


let rec apply_ignore_directive pl =
  match pl with
      `Ignore :: _ :: rest ->
	apply_ignore_directive rest
    | directive :: rest ->
	directive :: apply_ignore_directive rest
    | [] ->
	[]
;;


let extract_package_name pl =
  match 
    List.find (function `Name n -> true | _ -> false) pl
  with
      `Name n -> n
    | _ -> assert false
;;


let print_inst_plist file plist =
  let f = open_out file in
  try
    List.iter
      (function
	   `Name name ->
	     fprintf f "@name %s\n" name
	 | `Cwd dir ->
	     fprintf f "@cwd %s\n" dir
	 | `Src dir ->
	     fprintf f "@src %s\n" dir
	 | `Pkgdep d ->
	     fprintf f "@pkgdep %s\n" (print_dependency d)
	 | `Blddep d ->
	     fprintf f "@blddep %s\n" (print_dependency d)
	 | `Pkgcfl d ->
	     fprintf f "@pkgcfl %s\n" (print_dependency d)
	 | `Dirrm dir ->
	     fprintf f "@dirrm %s\n" dir
	 | `File(name,None) ->
	     fprintf f "%s\n" name
	 | `File(name,Some cs) ->
	     fprintf f "%s\n@comment %s\n" name cs
	 | `Comment s ->
	     fprintf f "@comment %s\n" s
	 | `Exec s ->
	     fprintf f "@exec %s\n" s
	 | `Unexec s ->
	     fprintf f "@unexec %s\n" s
	 | `Display file ->
	     fprintf f "@display %s\n" file
	 | `Option `Extract_in_place ->
	     fprintf f "@option extract-in-place\n"
	 | `Option `Preserve ->
	     fprintf f "@option preserve\n"
	 | `Ignore ->
	     fprintf f "@ignore\n"
	 | `Unsupported(directive,"") ->
	     fprintf f "%s\n" directive
	 | `Unsupported(directive,arg) ->
	     fprintf f "%s %s\n" directive arg
      )
      plist;
    close_out f
  with
      error ->
	close_out f; raise error
;;
